angular.module('deDrawing', ['toaster', 'deActions'])
	.factory('$deDrawing', ['$filter', '$rootScope', '$compile', function ($filter, $rootScope, $compile) {
		if (!$rootScope.temp) $rootScope.temp = {};
		var $deDrawing = {
		};
		return $deDrawing;
	}])
	.config(['$compileProvider', function ($compileProvider) {
		$compileProvider.debugInfoEnabled(false);
	}])
	.directive('deDrawing', ['$deDrawing', '$de', function ($deDrawing, $de) {
		return {
			//replace: true,
			scope: {
				data: '=deDrawing',
				actions: '=',
				save: '&',
			},
			templateUrl: $de.templatesFolder + 'drawing.htm',
			controller: ['$scope', '$attrs', '$element', '$deDrawing', '$timeout', '$compile', 'toaster', '$de', function ($scope, $attrs, $element, $deDrawing, $timeout, $compile, toaster, $de) {
				$scope.$deDrawing = $deDrawing;
				$scope.state = { mode: 'default', hoverHandle: new Handle() };

				$scope.state.mouse = { x: 0, y: 0 };
				$scope.state.selected = [];
				$scope.state.snapOffset = { x: 10, y: 10 };
				$scope.state.showgrid = true;
				$scope.state.connectmode = 'oneway';
				$scope.state.drawShape = 'rect';
				$scope.state.stroke = '#000000';
				$scope.state.fill = '#d7f2fc';
				$scope.shapes = $scope.data.shapes;
				$scope.shapesMap = {}
				$scope.undoLogs = [];
				$scope.redoLogs = [];

				angular.forEach($scope.shapes, function (shape) { if (shape.selected) $scope.state.selected.push(shape); })
				angular.forEach($scope.shapes, function (shape) { $scope.shapesMap[shape.id] = shape; });

				$scope.addShape = function (shape) {
					$scope.shapes.push(shape);
					$scope.shapesMap[shape.id] = shape;
				}

				$scope.addConnector = function (shape, connector) {
					shape.connectors.starting.push(connector.id);
				}

				$scope.removeConnector = function (connector) {
					delete $scope.shapesMap[connector.id];
					$de.remove($scope.shapes, connector);

					var shape = $scope.shapesMap[connector.start.shapeid];
					$de.remove(shape.connectors.starting, connector.id);
					$de.remove(shape.connectors.ending, connector.id);

					var shape = $scope.shapesMap[connector.end.shapeid];
					$de.remove(shape.connectors.starting, connector.id);
					$de.remove(shape.connectors.ending, connector.id);
				}

				$scope.removeShape = function (shape) {
					delete $scope.shapesMap[shape.id];
					$de.remove($scope.shapes, shape);
				}

				$scope.deleteShapes = function (shapes) {
					var shapesToDelete = [];
					angular.forEach(shapes, function (shape) {
						$de.add(shapesToDelete, shape);
						if( shape.shape != 'connector'){
							angular.forEach(shape.connectors.starting, function (connectorid) {
								var connector = $scope.shapesMap[connectorid];
								if (connector)
									$de.add(shapesToDelete, connector);
							})
							angular.forEach(shape.connectors.ending, function (connectorid) {
								var connector = $scope.shapesMap[connectorid];
								if (connector)
									$de.add(shapesToDelete, connector);
							})
						}
						else
							$scope.removeConnector(shape);
					})
					angular.forEach(shapesToDelete, function (shape) { $scope.removeShape(shape); })
					$scope.logUndo('deleted', shapesToDelete);
				}

				$scope.takeAction = function ($event, action, data) {
					action.onclick($event, action, data);
				}

				$scope.selectShape = function (shape) {
					shape.selected = true;
					Shape.resetOrig(shape);
					if ($scope.state.selected.indexOf(shape) < 0)
						$scope.state.selected.push(shape);
				}

				$scope.toggleSelectShape = function (shape) {
					shape.selected = !shape.selected;
					Shape.resetOrig(shape);
					var found = $scope.state.selected.indexOf(shape) >= 0;
					if (shape.selected && !found) {
						$scope.state.selected.push(shape);
					}
					else if (!shape.selected && found) {
						$de.remove($scope.state.selected, shape);
					}
				}

				$scope.unselectAllShapes = function () {
					$scope.state.selected = [];
					angular.forEach($scope.shapes, function (shape) {
						shape.selected = false;
					})
				}

				$scope.selectAllShapes = function (shape) {
					$scope.state.selected = [];
					angular.forEach($scope.shapes, function (shape) {
						shape.selected = true;
						Shape.resetOrig(shape);
						$scope.state.selected.push(shape);
					})
				}

				$scope.drawShape = function ($event, action, data) {
					lastShape = $scope.actionShape;
					$scope.state.mode = $scope.state.drawShape = action.action;
					switch (action.action) {
						case 'rect': lastShape.icon = 'fa-square-o'; break;
						case 'ellipse': lastShape.icon = 'fa-circle-o'; break;
						case 'diamond': lastShape.icon = 'fa-diamond'; break;
					}
					lastShape.action = action.action;
				}

				$scope.selectLineStyle = function ($event, action, data) {
					var lastStyle = $scope.actionLineType;
					//$scope.state.mode = $scope.state.drawShape = action.action;
					switch (action.action) {
						case 'connect-oneway': lastStyle.icon = 'fa-long-arrow-right'; $scope.state.connectmode = 'oneway'; break;
						case 'connect-twoway': lastStyle.icon = 'fa-exchange'; $scope.state.connectmode = 'twoway'; break;
						case 'connect-none': lastStyle.icon = 'fa-minus'; $scope.state.connectmode = ''; break;
					}
					var shapesToLog = [];
					angular.forEach($scope.state.selected, function (shape) {
						if (shape.shape == 'connector') {
							Shape.setProperties(shape, { connect: $scope.state.connectmode });
							shapesToLog.push(shape);
						}
					})
					$scope.logUndo('setproperties', shapesToLog);
					lastStyle.action = action.action;
				}

				$scope.drawingactions = [
					{
						icon: 'fa-mouse-pointer',
						action: 'move-resize',
						group: 'mode',
						active: true,
						onclick: function () {
							$scope.state.mode = 'default';
						}
					},
					{
						icon: 'fa-th',
						action: 'showgrid',
						active: $scope.state.showgrid,
						onclick: function ($event, action, data) {
							action.active = !action.active;
							$scope.state.showgrid = action.active;
						}
					},
					{
						icon: 'fa-square-o',
						group: 'mode',
						action: $scope.state.drawShape,
						onclick: $scope.drawShape,
						actions: [
							{
								icon: 'fa-square-o',
								action: 'rect',
								label: 'rectangle',
								onclick: $scope.drawShape,
							},
							{
								icon: 'fa-diamond',
								action: 'diamond',
								label: 'diamond',
								onclick: $scope.drawShape,
							},
							{
								icon: 'fa-circle-o',
								action: 'ellipse',
								label: 'ellipse',
								onclick: $scope.drawShape,
							}
						]
					},
					{
						icon: 'fa-mail-reply',
						action: 'undo',
						disabled: true,
						onclick: function ($event, action, data) {
							$scope.undo();
						}
					},
					{
						icon: 'fa-mail-forward',
						action: 'redo',
						disabled: true,
						onclick: function ($event, action, data) {
							$scope.redo();
						}
					},
					{
						icon: 'fa-long-arrow-right',
						action: 'connect-oneway',
						onclick: $scope.selectLineStyle,
						actions: [
								{
									icon: 'fa-long-arrow-right',
									label: 'One Way',
									action: 'connect-oneway',
									group: 'connectmode',
									onclick: $scope.selectLineStyle,
								},
								{
									icon: 'fa-exchange',
									action: 'connect-twoway',
									group: 'connectmode',
									label: 'Two Way',
									onclick: $scope.selectLineStyle,
								},
								{
									icon: 'fa-minus',
									action: 'connect-none',
									label: 'None',
									group: 'connectmode',
									onclick: $scope.selectLineStyle,
								}
						]
					},
					{
						style: { color: $scope.state.stroke },
						customicon: '<input type="color" ng-model="action.style.color" style="width:20px;height: 20px;" ng-click="$event.stopPropagation();" ng-change="takeAction($event,action)">',
						customlabel: '<font ng-style="action.style">A</font>',
						action: 'stroke',
						onclick: function ($event, action, data) {
							$scope.state.stroke = action.style.color;
						}
					},
					{
						style: { 'background-color': $scope.state.fill, width: '20px', height: '20px', border: '1px solid black', display: 'inline-block' },
						customicon: '<input type="color" ng-model="action.style[\'background-color\']" style="width:20px;height: 20px;" ng-click="$event.stopPropagation();" ng-change="takeAction($event,action)">',
						customlabel: '<div ng-style="action.style">&nbsp;</div>',
						action: 'fill',
						onclick: function ($event, action, data) {
							$scope.state.fill = action.style['background-color'];
						}
					},
				]

				$scope.actionShape = jQuery.grep($scope.drawingactions, function (action) { return action.action == $scope.state.drawShape; })[0];
				$scope.actionUndo = jQuery.grep($scope.drawingactions, function (action) { return action.action == 'undo'; })[0];
				$scope.actionRedo = jQuery.grep($scope.drawingactions, function (action) { return action.action == 'redo'; })[0];
				$scope.actionLineType = jQuery.grep($scope.drawingactions, function (action) { return action.action == 'connect-oneway'; })[0];

				$scope.logUndo = function (action, shapes) {
					var old = angular.copy(shapes);
					$scope.undoLogs.push({ action: action, shapes: shapes, old: old });
					delete $scope.redoLogs;
					$scope.redoLogs = [];
					$scope.actionUndo.disabled = false;
					$scope.actionRedo.disabled = true;
				}

				$scope.undo = function () {
					if ($scope.undoLogs.length > 0) {
						var action = $scope.undoLogs[$scope.undoLogs.length - 1];
						switch (action.action) {
							case 'added': angular.forEach(action.shapes, function (shape) { $scope.removeShape(shape); }); break;
							case 'moved':
							case 'resized': angular.forEach(action.shapes, function (shape, index) { Shape.resetToOrig(shape, action.old[index]); }); break;
							case 'deleted': angular.forEach(action.shapes, function (shape) { $scope.addShape(shape); }); break;
							case 'setproperties': angular.forEach(action.shapes, function (shape, index) { Shape.resetProperties(shape, action.old[index]); }); break;
						}
						$de.remove($scope.undoLogs, action);
						$scope.redoLogs.push(action);
						$scope.actionUndo.disabled = $scope.undoLogs.length <= 0;
						$scope.actionRedo.disabled = false;
					}
				}

				$scope.redo = function () {
					if ($scope.redoLogs.length > 0) {
						var action = $scope.redoLogs[$scope.redoLogs.length - 1];
						switch (action.action) {
							case 'added': angular.forEach(action.shapes, function (shape) { $scope.addShape(shape); }); break;
							case 'moved':
							case 'resized': angular.forEach(action.shapes, function (shape, index) { Shape.copyPosSize(shape, action.old[index]); }); break;
							case 'deleted': angular.forEach(action.shapes, function (shape) { $scope.removeShape(shape); }); break;
							case 'setproperties': angular.forEach(action.shapes, function (shape, index) { Shape.copyProperties(shape, action.old[index], action.old[index].originalProperties); }); break;
						}
						$de.remove($scope.redoLogs, action);
						$scope.undoLogs.push(action);
						$scope.actionUndo.disabled = false;
						$scope.actionRedo.disabled = $scope.redoLogs.length <= 0;
					}
				}

				$scope.resetState = function () {
					$scope.state.shapeadded = false;
					$scope.state.moveselected = false;
					$scope.state.connecting = false;
					$scope.state.currentShape = null;
					$scope.state.currentHandle = null;
				}

				setTimeout(function () {
					var svg = $element.find('svg');
					$scope.offset = svg.offset();
					svg.bind('mousemove', function (e) {
						//elements in foreignobject do not send correct mouseover event as they are treated as different body
						//hence we have to do manual calculation of any element inside forign object with respect to position of the
						//foreign object in the svg.
						var foreign = $(e.target).parents('foreignobject');
						var mouse = {};
						if (!foreign || !foreign.length) {
							mouse.x = e.clientX - $scope.offset.left;//e.offsetX;
							mouse.y = e.clientY - $scope.offset.top;//e.offsetY;
						}
						else {
							mouse.x = e.offsetX + parseInt(foreign.attr('x'));
							mouse.y = e.offsetY + parseInt(foreign.attr('y'));
						}
						if (!e.shiftKey) {
							mouse.x = $scope.state.snapOffset.x * Math.floor(mouse.x / $scope.state.snapOffset.x);
							mouse.y = $scope.state.snapOffset.y * Math.floor(mouse.y / $scope.state.snapOffset.y);
						}
						if ($scope.state.mouse.x == mouse.x && $scope.state.mouse.y == mouse.y) {
							//Mouse has not moved from previous position as per the snap to grid, so don't take any action
							//Not taking action on this stage may improve the performance.
						}
						else {
							$scope.state.mouse.x = mouse.x;
							$scope.state.mouse.y = mouse.y;
							//console.debug('x : ' + $scope.state.mouse.x + ' y : ' + $scope.state.mouse.y)
							//console.debug(e);
							if ($scope.state.currentShape && $scope.state.currentHandle) {
								if ($scope.state.currentHandle == 'move')
									$scope.state.cancelClick = true;
								var offset = { x: $scope.state.mouse.x - $scope.state.mousedown.x, y: $scope.state.mouse.y - $scope.state.mousedown.y }
								if ($scope.state.currentHandle.indexOf('connector') >= 0) {
									Shape.moveHandle($scope.state.currentShape, $scope.state.currentHandle, offset, $scope.state.mouse, $scope.state.mousedown);
								}
								else {
									if ($scope.state.currentShape.shape == 'selector') {
										Shape.moveHandle($scope.state.selector, $scope.state.currentHandle, offset, $scope.state.mouse, $scope.state.mousedown);
									}
									else if ($scope.state.moveselected) {
										angular.forEach($scope.state.selected, function (shape) {
											//shape.handle = $scope.state.currentShape.handle;
											Shape.moveHandle(shape, $scope.state.currentHandle, offset, $scope.state.mouse, $scope.state.mousedown);
										});
									}
									else {
										Shape.moveHandle($scope.state.currentShape, $scope.state.currentHandle, offset, $scope.state.mouse, $scope.state.mousedown);
									}
								}
								$scope.$apply();
							}
						}
					});

					$(document.body).bind('keydown', function (e) {
						console.debug(e);
						if (e.target.nodeName != 'TEXTAREA' && e.target.nodeName != 'INPUT') {
							var preventDefault = false;
							switch (e.keyCode) {
								case 37:// West
									angular.forEach($scope.state.selected, function (shape) {
										preventDefault = true;
										Shape.resetOrig(shape);
										Shape.moveHandle(shape, 'move', { x: -1 * $scope.state.snapOffset.x, y: 0 });
									});
									$scope.logUndo('moved', $scope.state.selected);
									break;
								case 38:// North
									angular.forEach($scope.state.selected, function (shape) {
										preventDefault = true;
										Shape.resetOrig(shape);
										Shape.moveHandle(shape, 'move', { x: 0, y: -1 * $scope.state.snapOffset.y });
									});
									$scope.logUndo('moved', $scope.state.selected);
									break;
								case 39:// East
									angular.forEach($scope.state.selected, function (shape) {
										preventDefault = true;
										Shape.resetOrig(shape);
										Shape.moveHandle(shape, 'move', { x: $scope.state.snapOffset.x, y: 0 });
									});
									$scope.logUndo('moved', $scope.state.selected);
									break;
								case 40:// South
									angular.forEach($scope.state.selected, function (shape) {
										preventDefault = true;
										Shape.resetOrig(shape);
										Shape.moveHandle(shape, 'move', { x: 0, y: $scope.state.snapOffset.y });
									});
									$scope.logUndo('moved', $scope.state.selected);
									break;
								case 65:
									if (e.ctrlKey) {
										preventDefault = true;
										$scope.selectAllShapes();
									}
									break;
								case 83: //'s' key
									if (e.ctrlKey) {
										preventDefault = true;
										$scope.save();
									}
									break;
								case 89: //'y' key
									if (e.ctrlKey)
										$scope.redo();
									break;
								case 90: //'z' key
									if (e.ctrlKey) {
										preventDefault = true;
										if (e.shiftKey)
											$scope.redo();
										else
											$scope.undo();
									}
									break;
								case 27: //Escape key
									if ($scope.state.shapeadded) {
										$scope.removeShape($scope.state.currentShape);
										$scope.resetState();
									}
									break;
								case 46: //delete key
									$scope.deleteShapes($scope.state.selected);
									preventDefault = true;
									$scope.unselectAllShapes();
									break;
								default:
									return;
							}
							if (preventDefault) {
								e.preventDefault();
								//e.stopPropagation();
							}
							$scope.$apply();
						}
					});

					svg.bind('mouseup', function (e) {
						if ($scope.state.currentShape) {
							if ($scope.state.currentShape.shape == 'selector') {
								if (!e.ctrlKey) {
									$scope.unselectAllShapes();
								}
								var startX = $scope.state.selector.left;
								var startY = $scope.state.selector.top;
								var endX = $scope.state.selector.left + $scope.state.selector.width;
								var endY = $scope.state.selector.top + $scope.state.selector.height;
								angular.forEach($scope.shapes, function (shape) {
									var shapeStart = Shape.getStart(shape, $scope.shapesMap);
									var shapeEnd = Shape.getEnd(shape, $scope.shapesMap);
									//console.debug('start : ' + startX + ' ' + startY + ' end : ' + endX + ' ' + endY);
									//console.debug('SHAPE : start : ' + shapeStart.x + ' ' + shapeStart.y + ' end : ' + shapeEnd.x + ' ' + shapeEnd.y);
									if ((shapeStart.x <= endX && shapeStart.x >= startX) && (shapeStart.y <= endY && shapeStart.y >= startY) &&
										(shapeEnd.x <= endX && shapeEnd.x >= startX) && (shapeEnd.y <= endY && shapeEnd.y >= startY)) {
										$scope.selectShape(shape);
									}
									shape.$state.editing = false;
								})
								$scope.state.selector = null;
								$scope.state.currentHandle = null;
							}
							else {
								console.log('up : ' + $scope.state.currentHandle);
								switch ($scope.state.currentHandle) {
									case 'move':
										if ($scope.state.mouse.x != $scope.state.mousedown.x || $scope.state.mouse.y != $scope.state.mousedown.y)
											$scope.logUndo('moved', ($scope.state.moveselected) ? $scope.state.selected : [$scope.state.currentShape]);
										break;
									case 'west':
									case 'north':
									case 'east':
									case 'south':
									case 'north-west':
									case 'north-east':
									case 'south-west':
									case 'south-east':
										if ($scope.state.mouse.x != $scope.state.mousedown.x || $scope.state.mouse.y != $scope.state.mousedown.y) {
											if( !$scope.state.shapeadded )
												$scope.logUndo('resized', ($scope.state.moveselected) ? $scope.state.selected : [$scope.state.currentShape]);
										}
										else if ($scope.state.shapeadded) {
											$scope.removeShape($scope.state.currentShape);
										}
										break;
									case 'connector-start':
										if ($scope.state.hoverHandle.shape) {
											if ($scope.state.hoverHandle.shape.id == $scope.state.currentShape.end.shapeid) {
												$scope.removeShape($scope.state.currentShape);
											}
											else {
												Shape.setConnectorStart($scope.state.currentShape, $scope.state.hoverHandle.shape, $scope.state.hoverHandle.handle);
											}
										}
										else
											$scope.removeShape($scope.state.currentShape);
										break;
									case 'connector-end':
										if ($scope.state.hoverHandle.shape) {
											if ($scope.state.hoverHandle.shape.id == $scope.state.currentShape.start.shapeid) {
												$scope.removeShape($scope.state.currentShape);
											}
											else {
												if ($scope.state.shapeadded)
													$scope.logUndo('added', [$scope.state.currentShape]);
												Shape.setConnectorEnd($scope.state.currentShape, $scope.state.hoverHandle.shape, $scope.state.hoverHandle.handle);
											}
										}
										else
											$scope.removeShape($scope.state.currentShape);
										break;
								}
							}
						}
						$scope.resetState();
						$scope.$apply();
					});

					svg.bind('mousedown', function (e) {
						$scope.state.mousedown = { x: $scope.state.mouse.x, y: $scope.state.mouse.y }
						switch ($scope.state.mode) {
							case 'rect':
							case 'ellipse':
							case 'diamond':
								$scope.unselectAllShapes();
								$scope.state.currentShape = Shape.create($scope.state.drawShape, $scope.state.mouse.x, $scope.state.mouse.y, 0, 0, $scope.state.stroke, $scope.state.fill);
								$scope.state.currentHandle = 'south-east';
								Shape.resetOrig($scope.state.currentShape);
								$scope.addShape($scope.state.currentShape);
								$scope.state.moveselected = false;
								$scope.state.shapeadded = true;
								$scope.logUndo('added', [$scope.state.currentShape]);
								//e.preventDefault();
								break;
							default:
								if ($(e.target).hasClass('background')) {
									$scope.state.currentShape =
									$scope.state.selector = Shape.create('selector', $scope.state.mouse.x, $scope.state.mouse.y, 0, 0);
									$scope.state.currentHandle = 'south-east';
									Shape.resetOrig($scope.state.currentShape);
									e.preventDefault();
								}
								break;
						}
					});
				}, 300);

				$scope.handle = function (shape, event) {
					$scope.state.mousedown = { x: $scope.state.mouse.x, y: $scope.state.mouse.y }
					console.log(shape);
					$scope.state.currentShape = shape;

					if ($(event.currentTarget).hasClass('connect-point')) {
						var connector = Shape.createConnector(shape.id, $scope.state.currentHandle, event.offsetX, event.offsetY, $scope.state.stroke, $scope.state.fill);
						Shape.setConnectorStart(connector, shape, $scope.state.currentHandle);
						connector.connect = $scope.state.connectmode;
						$scope.addShape(connector);
						$scope.state.currentHandle = 'connector-end';
						$scope.state.currentShape = connector;
						$scope.state.shapeadded = true;
					}

					switch ($scope.state.currentHandle) {
						case 'selected':
							if (!$scope.state.cancelClick) {
								if (event.ctrlKey) {
									$scope.toggleSelectShape(shape);
								}
								else {
									$scope.unselectAllShapes();
									$scope.selectShape(shape);
								}
							}
							else
								$scope.state.cancelClick = false;
							$scope.state.currentHandle = null;
							break;
						case 'connector-start':
							$scope.state.currentShape.start.shapeid = null;
							$scope.state.currentShape.start.handle = null;
							$scope.state.currentShape.start.x = event.offsetX;
							$scope.state.currentShape.start.y = event.offsetY;
							$scope.state.connecting = true;
							break;
						case 'connector-end':
							$scope.state.currentShape.end.shapeid = null;
							$scope.state.currentShape.end.handle = null;
							$scope.state.currentShape.end.x = event.offsetX;
							$scope.state.currentShape.end.y = event.offsetY;
							$scope.state.connecting = true;
							break;
						default: break;
					}
				}
			}]
		}
	}])
.directive('deShape', ['$deDrawing', function ($deDrawing) {
	return {
		replace: true,
		scope: {
			shapes: '=',
			state: '=',
			shape: '=deShape',
			handle: '&',
		},
		template: '<g ng-class="{\'selected\': shape.selected, \'hover\': hover}" ng-include="templateUrl" class="shape" />',
		//template: '<rect ng-attr-x="{{shape.left}}" />',
		controller: ['$scope', '$attrs', '$element', '$deDrawing', '$timeout', '$compile', '$de', function ($scope, $attrs, $element, $deDrawing, $timeout, $compile, $de) {
			$scope.$deDrawing = $deDrawing;
			$scope.Math = Math;
			$scope.shape.$state = { editing: false };

			$scope.selecthandle = function ($event, handle) {
				if (!handle)
					handle = $($event.currentTarget).attr('handle');
				if (handle) {
					$scope.state.currentHandle = handle;
					$scope.state.moveselected = $scope.shape.selected;
					Shape.resetOrig($scope.shape);
					console.debug('down: ' + handle);
					$scope.handle({ shape: $scope.shape, event: $event });
				}
				$event.stopPropagation();
				$event.preventDefault();
			}

			$scope.donothing = function ($event) {
				$event.stopPropagation();
			}

			$scope.edit = function ($event) {
				$scope.shape.$state.editing = true;
				setTimeout(function () { $element.find('textarea').focus() }, 50);
			}

			$scope.stopedit = function ($event) {
				$scope.shape.$state.editing = false;
			}

			$scope.select = function ($event) {
				console.debug($scope.shape.id + ' : selected');
				$scope.state.currentHandle = 'selected';
				Shape.resetOrig($scope.shape);
				$scope.handle({ shape: $scope.shape, event: $event });
				$event.stopPropagation();
			}

			$scope.stopPropagation = function ($event) {
				$event.stopPropagation();
			}

			$scope.handleclicked = function ($event) {
				$event.stopPropagation();
			}

			if ($scope.shape.shape != 'connector') {
				setTimeout(function () {
					$element.next().mouseenter(function () {
						$scope.hover = true;
						$element.addClass('hover');
					}).mouseleave(function () {
						$scope.hover = false;
						$element.removeClass('hover');
					});
					
					$element.next().find('.connect-point').mouseenter(function (e) {
						$(e.currentTarget).addClass('hover');
						$scope.state.hoverHandle.shape = $scope.shape;
						$scope.state.hoverHandle.handle = $(e.currentTarget).attr('handle');
					}).mouseout(function (e) {
						$(e.currentTarget).removeClass('hover');
					})
				}, 300)
			}

			$scope.path = function () {
				var startHandle = $scope.shape.start.handle, endHandle = $scope.shape.end.handle;
				var startShape = $scope.shapes[$scope.shape.start.shapeid];
				var endShape = $scope.shapes[$scope.shape.end.shapeid];
				var start = Shape.getStart($scope.shape, $scope.shapes);
				var end = Shape.getEnd($scope.shape, $scope.shapes);

				var path = 'M ' + start.x + ' ' + start.y + ' L ' + end.x + ' ' + end.y;
				$scope.endarrow = '';
				var offset = { x: $scope.state.snapOffset.x * 2, y: $scope.state.snapOffset.y * 2 };
				var negativeOffset = { x: offset.x * -1, y: offset.y * -1 };

				if (startHandle && endHandle) {
					if (startHandle == 'west' && endHandle == 'west') {
						if (start.x > end.x && start.y > (endShape.top - offset.y) && start.y < (endShape.bottom + offset.y))
							path = Path.calcWestWest(end, start, endShape, startShape);
						else
							path = Path.calcWestWest(start, end, startShape, endShape);
					}
					else if (startHandle == 'west' && endHandle == 'north') {
						path = Path.calcWestNorth(start, end, startShape, endShape);
					}
					else if (startHandle == 'west' && endHandle == 'east') {
						path = Path.calcWestEast(start, end, startShape, endShape);
					}
					else if (startHandle == 'west' && endHandle == 'south') {
						if (start.y > end.y) {
							path = 'M ' + start.x + ' ' + start.y +
									' h ' + (end.x - start.x) +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							var middle = start.x - $scope.state.snapOffset.x * 2;
							var top = end.y + $scope.state.snapOffset.y * 2;
							path = 'M ' + start.x + ' ' + start.y +
									' H ' + middle + ' V ' + top + ' H ' + end.x +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'north' && endHandle == 'west') {
						path = Path.calcWestNorth(end, start, endShape, startShape);
					}
					else if (startHandle == 'north' && endHandle == 'north') {
						var top = Math.min(start.y, end.y) - $scope.state.snapOffset.y * 2;
						path = 'M ' + start.x + ' ' + start.y +
								' V ' + top + ' h ' + (end.x - start.x) +
								' L ' + end.x + ' ' + end.y;
					}
					else if (startHandle == 'north' && endHandle == 'east') {
						if (start.y > end.y) {
							path = 'M ' + start.x + ' ' + start.y +
									' v ' + (end.y - start.y) +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							var middle = end.x + $scope.state.snapOffset.y * 2;
							var top = start.y - $scope.state.snapOffset.y * 2;
							path = 'M ' + start.x + ' ' + start.y +
									' V ' + top + ' H ' + middle + ' V ' + end.y +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'north' && endHandle == 'south') {
						if (start.y > end.y) {
							path = 'M ' + start.x + ' ' + start.y +
									' v ' + ((end.y - start.y) / 2) + ' h ' + (end.x - start.x) +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							path = 'M ' + start.x + ' ' + start.y +
									' v ' + (-1 * $scope.state.snapOffset.y * 2) + ' H ' + (Math.max(startShape.right, endShape.right) + $scope.state.snapOffset.x * 2) +
									' V ' + (end.y + $scope.state.snapOffset.y * 2) + ' H ' + end.x +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'east' && endHandle == 'west') {
						if (start.x < end.x) {
							path = 'M ' + start.x + ' ' + start.y +
									' h ' + ((end.x - start.x) / 2) + ' v ' + (end.y - start.y) +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							var top = Math.max(startShape.bottom, endShape.bottom) + $scope.state.snapOffset.y * 2;
							path = 'M ' + start.x + ' ' + start.y +
									' H ' + (Math.max(start.x, endShape.right) + $scope.state.snapOffset.x * 2) + ' V ' + top + ' H ' + (end.x - $scope.state.snapOffset.x * 2) +
									' V ' + end.y +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'east' && endHandle == 'north') {
						if (start.y < end.y) {
							path = 'M ' + start.x + ' ' + start.y +
									' h ' + (end.x - start.x) +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							var middle = start.x + $scope.state.snapOffset.x * 2;
							var top = end.y - $scope.state.snapOffset.y * 2;
							path = 'M ' + start.x + ' ' + start.y +
									' H ' + middle + ' V ' + top + ' H ' + end.x +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'east' && endHandle == 'east') {
						var right = Math.max(start.x, end.x) + $scope.state.snapOffset.x * 2;
						path = 'M ' + start.x + ' ' + start.y +
								' H ' + right + ' v ' + (end.y - start.y) +
								' L ' + end.x + ' ' + end.y;
					}
					else if (startHandle == 'east' && endHandle == 'south') {
						if (start.y < end.y) {
							var middle = start.x + $scope.state.snapOffset.x * 2;
							var top = end.y + $scope.state.snapOffset.y * 2;
							path = 'M ' + start.x + ' ' + start.y +
									' H ' + middle + ' V ' + top + ' H ' + end.x +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							path = 'M ' + start.x + ' ' + start.y +
									' h ' + (end.x - start.x) +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'south' && endHandle == 'west') {
						if (start.y < end.y) {
							path = 'M ' + start.x + ' ' + start.y +
									' v ' + (end.y - start.y) +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							var middle = end.x - $scope.state.snapOffset.y * 2;
							var top = start.y + $scope.state.snapOffset.y * 2;
							path = 'M ' + start.x + ' ' + start.y +
									' V ' + top + ' H ' + middle + ' V ' + end.y +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'south' && endHandle == 'north') {
						path = 'M ' + start.x + ' ' + start.y +
								' v ' + ((end.y - start.y) / 2) + ' h ' + (end.x - start.x) +
								' L ' + end.x + ' ' + end.y;
					}
					else if (startHandle == 'south' && endHandle == 'east') {
						if ((start.x > end.x) && (start.y < end.y)) {
							path = 'M ' + start.x + ' ' + start.y +
									' v ' + (end.y - start.y) +
									' L ' + end.x + ' ' + end.y;
						}
						else {
							path = 'M ' + start.x + ' ' + start.y +
									' V ' + (Math.max(startShape.bottom, endShape.bottom) + $scope.state.snapOffset.y * 2) + ' H ' + (Math.max(startShape.right, endShape.right) + $scope.state.snapOffset.y * 2) + ' V ' + end.y +
									' L ' + end.x + ' ' + end.y;
						}
					}
					else if (startHandle == 'south' && endHandle == 'south') {
						var bottom = Math.max(start.y, end.y) + $scope.state.snapOffset.y * 2;
						path = 'M ' + start.x + ' ' + start.y +
								' V ' + bottom + ' h ' + (end.x - start.x) +
								' L ' + end.x + ' ' + end.y;
					}

					$scope.start = start;
					$scope.end = end;

					$scope.endarrow = Path.getArrow(end, endHandle);
					$scope.startarrow = Path.getArrow(start, startHandle);
				}

				$scope.connectorPath = path;
				return path;
			}


			console.debug($scope.shape);
			$scope.templatesFolder = $de.templatesFolder + 'shapes/';
			$scope.templateUrl = $scope.templatesFolder + $scope.shape.shape + '.htm';
			$scope.textContentTemplateUrl = $scope.templatesFolder + 'text-content.htm';
			$scope.handlesTemplateUrl = $scope.templatesFolder + 'handles.htm';
		}]
	}
}])

var Path = {
	offset: { x: 20, y: 20 },
	negativeOffet: { x: -20, y: -20 },
	calcWestWest: function (start, end, startShape, endShape) {
		var v;
		if (start.y <= end.y) {
			v = Math.max(end.y, startShape.bottom + this.offset.y);
		}
		else {
			v = Math.min(end.y, startShape.top - this.offset.y);
		}

		var path = 'M ' + start.x + ' ' + start.y +
				' H ' + (Math.min(start.x, end.x) - this.offset.x) +
				' V ' + v +
				' H ' + (end.x - this.offset.x) +
				' V ' + end.y +
				' L ' + end.x + ' ' + end.y;
		return path;
	},
	calcWestNorth: function (start, end, startShape, endShape) {
		var v, h;
		if ((startShape.bottom + Path.offset.y) >= end.y && end.x > startShape.left)
			v = Math.min(startShape.top, endShape.top) - Path.offset.y;
		else
			v = (end.x < startShape.left) ? end.y - Path.offset.y : startShape.bottom + Path.offset.y;

		if (end.y > start.y && end.x < (start.x))
			h = end.x;
		else if (end.y < start.y && endShape.left < start.x && endShape.right >= (start.x - Path.offset.x))
			h = endShape.left - Path.offset.x;
		else
			h = start.x - Path.offset.x;

		var path = 'M ' + start.x + ' ' + start.y +
					' H ' + h +
					' V ' + v +
					' H ' + end.x +
					' L ' + end.x + ' ' + end.y;
		return path;
	},
	calcWestEast: function (start, end, startShape, endShape) {
		var v, h;
		if (end.x > start.x - 2 * Path.offset.x && endShape.left < start.x - Path.offset.x)
			h = endShape.left - Path.offset.x;
		else
			h = start.x - Path.offset.x;
		if (end.x <= start.x - 2 * Path.offset.x && end.y >= start.y)
			v = end.y;
		else if (start.y > end.y)
			v = Math.max(startShape.bottom, endShape.bottom) + Path.offset.y;
		else
			v = Math.min(startShape.top, endShape.top) - Path.offset.x;
		var path = 'M ' + start.x + ' ' + start.y +
				' H ' + h +
				' V ' + v +
				' H ' + (end.x + Path.offset.x) +
				' V ' + end.y +
				' L ' + end.x + ' ' + end.y;
		return path;
	},
	calcWestSouth: function (start, end, startShape, endShape) {
	},
	getArrow: function (point, direction) {
		var arrow;
		switch (direction) {
			case 'east': arrow = 'M ' + point.x + ' ' + point.y + ' l 8 3 v -6 Z'; break;
			case 'west': arrow = 'M ' + point.x + ' ' + point.y + ' l -8 3 v -6 Z'; break;
			case 'south': arrow = 'M ' + point.x + ' ' + point.y + ' l 3 8 h -6 Z'; break;
			case 'north': arrow = 'M ' + point.x + ' ' + point.y + ' l 3 -8 h -6 Z'; break;
		}
		return arrow;
	}
}

function Point(x, y) {
	this.x = x;
	this.y = y;
}

function guid() {
	function s4() {
		return Math.floor((1 + Math.random()) * 0x10000)
		  .toString(16)
		  .substring(1);
	}
	return s4() + s4() + '-' + s4() + '-' + s4() + '-' +
	  s4() + '-' + s4() + s4() + s4();
}

function Size(width, height) {
	this.width = width;
	this.height = height;
}

function Handle(shape, handle) {
	this.shapeid = shape;
	this.handle = handle;
}

var Shape = {
	create: function (shape, left, top, width, height, stroke, fill, text) {
		var shapeNew = new Object();
		shapeNew.id = guid();
		shapeNew.shape = shape;
		shapeNew.left = left;
		shapeNew.top = top;
		shapeNew.width = width;
		shapeNew.height = height;
		shapeNew.right = left + width;
		shapeNew.bottom = top + height;
		shapeNew.text = text;
		shapeNew.stroke = stroke;
		shapeNew.fill = fill;
		shapeNew.connectors = { starting: [], ending: [] };
		return shapeNew;
	},
	createConnector: function (startShape, startHandle, endShape, endHandle, stroke, fill) {
		var connectorNew = new Object();
		connectorNew.id = guid();
		connectorNew.shape = 'connector';
		connectorNew.stroke = stroke;
		connectorNew.fill = fill;
		if (typeof startShape === 'string')
			connectorNew.start = new Handle(startShape, startHandle);
		else
			connectorNew.start = new Point(startShape, startHandle);

		if (typeof endShape === 'string')
			connectorNew.end = new Handle(endShape, endHandle);
		else
			connectorNew.end = new Point(endShape, endHandle);

		return connectorNew;
	},
	setProperties: function (shape, properties) {
		shape.originalProperties = {};
		angular.forEach(properties, function (value, key) {
			shape.originalProperties[key] = shape[key];
			shape[key] = value;
		})
	},
	resetProperties: function (shape, old) {
		if (old.originalProperties) {
			angular.forEach(old.originalProperties, function (value, key) {
				shape[key] = old.originalProperties[key];
			})
		}
	},
	copyProperties: function (shape, old, properties) {
		if (properties) {
			angular.forEach(properties, function (value, key) {
				shape[key] = old[key];
			})
		}
	},
	setLeft: function (shape, left) {
		shape.left = left;
		shape.right = shape.left + shape.width;
	},
	setTop: function (shape, top) {
		shape.top = top;
		shape.bottom = shape.top + shape.height;
	},
	setConnectorStart: function (connector, shape, handle) {
		connector.start.shapeid = shape.id;
		shape.connectors.starting.push(connector.id);
		connector.start.handle = handle;
	},
	setConnectorEnd: function (connector, shape, handle) {
		connector.end.shapeid = shape.id;
		shape.connectors.ending.push(connector.id);
		connector.end.handle = handle;
	},
	getStart: function (shape, shapes) {
		var point;
		switch (shape.shape) {
			case 'connector':
				if (shape.start.shapeid) {
					var startShape = shapes[shape.start.shapeid];
					if (startShape) {
						point = Shape.getHandlePosition(startShape, shape.start.handle);
					}
				}
				else
					point = shape.start;
				break;
			default:
				point = new Point(shape.left, shape.top);
		}
		if (point)
			return point;
		else
			return new Point();
	},
	getEnd: function (shape, shapes) {
		var point;
		switch (shape.shape) {
			case 'connector':
				if (shape.end.shapeid) {
					var endShape = shapes[shape.end.shapeid];
					if (endShape) {
						point = Shape.getHandlePosition(endShape, shape.end.handle);
					}
				}
				else
					point = shape.end;
				break;
			default:
				point = new Point(shape.left + shape.width, shape.top + shape.height);
		}
		if (point)
			return point;
		else
			return new Point();
	},
	resetOrig: function (shape) {
		shape.originalPosition = { x: shape.left, y: shape.top };
		shape.originalSize = { width: shape.width, height: shape.height };
	},
	resetToOrig: function (shape, old) {
		shape.width = old.originalSize.width;
		shape.height = old.originalSize.height;
		Shape.setLeft(shape, old.originalPosition.x);
		Shape.setTop(shape, old.originalPosition.y);
	},
	copyPosSize: function (shape, old) {
		shape.width = old.width;
		shape.height = old.height;
		Shape.setLeft(shape, old.left);
		Shape.setTop(shape, old.top);
	},
	resizeEast: function (shape, offset) {
		shape.width = shape.originalSize.width + offset.x;
		if (shape.width < 0) {
			shape.width *= -1;
			shape.left = shape.originalPosition.x - shape.width;
		}
		shape.right = shape.left + shape.width;
	},
	resizeWest: function (shape, offset) {
		shape.left = shape.originalPosition.x + offset.x;
		shape.width = shape.originalSize.width - offset.x;
		if (shape.width < 0) {
			shape.width *= -1;
			shape.left = shape.originalPosition.x + shape.originalSize.width;
		}
		shape.right = shape.left + shape.width;
	},
	resizeNorth: function (shape, offset) {
		shape.top = shape.originalPosition.y + offset.y;
		shape.height = shape.originalSize.height - offset.y;
		if (shape.height < 0) {
			shape.height *= -1;
			shape.top = shape.originalPosition.y + shape.originalSize.height;
		}
		shape.bottom = shape.top + shape.height;
	},
	resizeSouth: function (shape, offset) {
		shape.height = shape.originalSize.height + offset.y;
		if (shape.height < 0) {
			shape.height *= -1;
			shape.top = shape.originalPosition.y - shape.height;
		}
		shape.bottom = shape.top + shape.height;
	},
	getHandlePosition: function (shape, handle) {
		var point = new Point();
		switch (handle) {
			case 'east':
				point.x = shape.left + shape.width;
				point.y = shape.top + shape.height / 2;
				break;
			case 'south':
				point.x = shape.left + shape.width / 2;
				point.y = shape.top + shape.height;
				break;
			case 'west':
				point.x = shape.left;
				point.y = shape.top + shape.height / 2;
				break;
			case 'north':
				point.x = shape.left + shape.width / 2;
				point.y = shape.top;
				break;
			case 'south-west':
				point.x = shape.left;
				point.y = shape.top + shape.height;
				break;
			case 'south-east':
				point.x = shape.left + shape.width;
				point.y = shape.top + shape.height;
				break;
			case 'north-west':
				point.x = shape.left;
				point.y = shape.top;
				break;
			case 'north-east':
				point.x = shape.left + shape.width;
				point.y = shape.top;
				break;
		}
		return point;
	},
	moveHandle: function (shape, handle, offset, point) {

		switch (handle) {
			case 'east': Shape.resizeEast(shape, offset); break;
			case 'south': Shape.resizeSouth(shape, offset); break;
			case 'west': Shape.resizeWest(shape, offset); break;
			case 'north': Shape.resizeNorth(shape, offset); break;
			case 'south-west': Shape.resizeSouth(shape, offset); Shape.resizeWest(shape, offset); break;
			case 'south-east': Shape.resizeSouth(shape, offset); Shape.resizeEast(shape, offset); break;
			case 'north-west': Shape.resizeNorth(shape, offset); Shape.resizeWest(shape, offset); break;
			case 'north-east': Shape.resizeNorth(shape, offset); Shape.resizeEast(shape, offset); break;
			case 'move':
				Shape.setLeft(shape, shape.originalPosition.x + offset.x);
				Shape.setTop(shape, shape.originalPosition.y + offset.y);
				//console.debug(point);
				//console.debug('left : ' + shape.left + ' top : ' + shape.top + ' width : ' + shape.width + ' height : ' + shape.height);
				break;
			case 'connector-start':
				shape.start.x = point.x;
				shape.start.y = point.y;
				break;
			case 'connector-end':
				shape.end.x = point.x;
				shape.end.y = point.y;
				break;
		}
	}
}